#include <stdlib.h>
#include <math.h>
#include <R.h>
#include <Rmath.h>        /* for 'runif' */
#include <R_ext/Utils.h>  /* for rsort_with_index */


/* Modified from entry in: 'R-2.4.0/include/R_ext/RS.h' 
#define Calloc(n, t)   (t *) calloc( (size_t) (n), sizeof(t) )
*/

/* Replace later(?) with boolean typed variables? */
/* Need to #include<stdbool.h> for that */
#define TRUE 1
#define FALSE 0


/* =========================================================================== */

int Sign(double x)
{
    return ((x > 0) ? 1 : ((x == 0)? 0 : -1));
}


/* =========================================================================== */

/* This function was taken from R-2.4.0/src/main/sort.c, modified  */
/* by replacing "rcmp(x[j - h],v,TRUE) > 0" with "(x[j - h] > v)"  */
/* then renamed (to avoid getting the R function when compiling) */

void indexSort(double *x, int *indx, int n)
{
    double v;
    int i, j, h, iv;

    for (h = 1; h <= n / 9; h = 3 * h + 1);
    for (; h > 0; h /= 3)
        for (i = h; i < n; i++) {
            v = x[i]; iv = indx[i];
            j = i;
            while (j >= h && (x[j - h] > v))
                 { x[j] = x[j - h]; indx[j] = indx[j-h]; j -= h; }
            x[j] = v; indx[j] = iv;
        }
}


/* =========================================================================== */

/* Tabulation functions */


void freq1D(int *x, int lev, int N, double *freqVec) 
{
    int i; 

    /* initialize region of tabVec that we'll be using to zero */
    for (i = 0; i < lev; i++)
        freqVec[i] = 0;

    /* level codes start at 1, so need to subtract 1 to start them at 0 */
    for (i = 0; i < N; i++)
        freqVec[ x[i]-1 ]++ ; 

    /* convert counts to freqs/probabilities */ 
    for (i = 0; i < lev; i++)
        freqVec[i] = freqVec[i]/N; 
}


void freq2D(int *x, int *y, int x_lev, int y_lev, int N, double **freqMat) 
{
    int i, j; 

    /* initialize the region of freqMat that we'll be using to zero */
    for (i = 0; i < x_lev; i++)
        for (j = 0; j < y_lev; j++)
            freqMat[i][j] = 0;

    /* level codes start at 1, so need to subtract 1 to start them at 0 */
    for (i = 0; i < N; i++)
        freqMat[ x[i]-1 ][ y[i]-1 ]++ ; 

    /* convert counts to freqs/probabilities */ 
    for (i = 0; i < x_lev; i++)
        for (j = 0; j < y_lev; j++)
            freqMat[i][j] = freqMat[i][j]/N;
}



/* =========================================================================== */



void calcStats(int **x, int Nseq, int Npos, int *posI, int *posJ,
               int Npairs, double **jntprobIJ, int *Nlev, int maxLvl,
	       double **wt, int useWt, double **p, double *sum_PlogP,
               double *M, double *Mstar, int *anyNAN)
{
    int pair, k, l;
    int pos_i, pos_j;
    double pIJ, norm_const;
    long double tmpVal;

   /* NOTE: Anything done here using wt.s needs to be conditional      */
   /*       on:  (useWt != 0).  If it's too hard/clumsy to incorporate */
   /*       wt in this fashion, perhaps write function 'calcWtStats'    */

    for (pair = 0; pair < Npairs; pair++)
    {
        /* Need to subtract one for vector indexing */
        pos_i = posI[pair] - 1;
        pos_j = posJ[pair] - 1;

 
	/* get jnt freq of pos_i and pos_j of sequences */ 
	freq2D(x[pos_i], x[pos_j], Nlev[pos_i], Nlev[pos_j],
               Nseq, jntprobIJ);

        tmpVal = 0; 
        for (k = 0; k < Nlev[pos_i]; k++)
           for (l = 0; l < Nlev[pos_j]; l++)
	   {
	       pIJ = jntprobIJ[k][l];
               if ( pIJ > 0 ) 
                   tmpVal +=  pIJ * log( pIJ / (p[pos_i][k] * p[pos_j][l]));
           }


        /* If norm_const=0 then Mstar will be 0/0 = NaN.  When
         * this occurs we note it by setting *anyNAN = TRUE */
        norm_const = sum_PlogP[pos_i] + sum_PlogP[pos_j];

        M[pair] = tmpVal; 
        Mstar[pair] = -2 * tmpVal / norm_const;

        /* Note: values of 'norm_const' are always <= zero */
        if (norm_const == 0)
            *anyNAN = TRUE;
    } 
}
/* End calcStats */


/* =========================================================================== */


void singleGroupPairs(int *xvec, int *ptNseq, int *ptNpos, int *Nlev, 
                      int *posI, int *posJ, int *ptNpairs,
                      double *wtvec, int *ptuseWt, double *pvec, int *ptNperm,
                      double *M, double *M_p, double *Mstar, double *Mstar_p)
{

    /* define variable types */
    int i, j, k, pair;
    int Nseq, Npos, Npairs, useWt, Nperm;
    long double tmpVal;
     
    Nseq = *ptNseq;
    Npos = *ptNpos;
    Npairs = *ptNpairs;
    useWt = *ptuseWt;
    Nperm = *ptNperm;


    /*** BEGIN HOUSEKEEPING ***/

    /* All matrices/list from R have to be passed to C as vectors, 
     * but for clarity, we'd like to access them as if they are matrices.
     * So we set up the infra-structure to allow us to do that. 
     */

    /* Create int matrix 'x' of dim Npos x Nseq to access data in 'xvec' */
    int **x;
    x = Calloc(Npos, int *);
    for (i = 0; i < Npos; i++)
         x[i] = & ( xvec[i*Nseq] );

    /* Create 'p', length Npos, to access marginal prob.s of each position */
    double **p;
    int loc;
    p = Calloc(Npos, double *);
    loc = 0;
    for (i = 0; i < Npos; i++)
    {
         p[i] = & ( pvec[loc] );
         loc += Nlev[i];
    }

    /* If useWt != 0, create Nseq x Nseq matrix 'wt' to access weights */
    double **wt;
    if (useWt != 0)
    {
        wt = Calloc(Nseq, double *);
        for (j = 0; j < Nseq; j++)
             wt[j] = & ( wtvec[j*Nseq] );
    }

    /* Next, we set up a matrix to be used in storing jnt freq results
     * for a pair of positions.  Then pre-calculate some quantities 
     * needed for computation of statistic Mstar. 
     */

    /* Find the maximum number of levels across all positions: maxLvl */
    int maxLvl;
    maxLvl = 0;
    for (i = 0; i < Npos; i++)
        if (Nlev[i] > maxLvl)
            maxLvl = Nlev[i]; 


    /* Create a matrix of dimension (maxLvl x maxLvl) to be used by
     * function 'calcStats' for storing jnt freqs of pos i and j  
     */ 
    double **jntprobIJ; 
    jntprobIJ = Calloc(maxLvl, double *);
    for (k = 0; k < maxLvl; k++)
        jntprobIJ[k] = Calloc(maxLvl, double);

    /* Pre-calculate terms needed for statistic Mstar */
    double *sum_PlogP;
    sum_PlogP = Calloc(Npos, double);
    for (i = 0; i < Npos; i++)
    {
       tmpVal = 0; 
       for (k = 0; k < Nlev[i]; k++)
           tmpVal += p[i][k] * log( p[i][k] );

       sum_PlogP[i] = tmpVal; 
    }
    /*** HOUSEKEEPING DONE ***/



    /*** NOW LETS DO SOME WORK! ***/

    /* define/initialize indicator of NAN occurrance in calcStats() */
    int anyNAN;
    anyNAN = FALSE;

    /* (1) calculate statistics using the original data */
    calcStats(x, Nseq, Npos, posI, posJ, Npairs, jntprobIJ, Nlev, maxLvl,
              wt, useWt, p, sum_PlogP, M, Mstar, &anyNAN);

    /* This code should never run.  Original data was screened to remove
     * remove positions with no variation in them. But just in case... */
    if (anyNAN == TRUE)
    {
        printf("%s", "Encountered unexpected problem w/original data:\n");
        printf("%s", "NaN values produced for Mstar in function",
                     " calcStats.  Exiting...\n\n");
        exit(123);
    }



    /* (2) Define/allocate/initialize storage for null data, null statistics,
     *     and for counts of null statistics more extreme than observed ones
     */ 
    int **null_x;
    null_x = Calloc(Npos, int *);
    for (i=0; i<Npos; i++) 
        null_x[i] = Calloc(Nseq, int);

    double *null_M, *null_Mstar;
    null_M = Calloc(Npairs, double);
    null_Mstar = Calloc(Npairs, double);

    int *null_M_cnt, *null_Mstar_cnt;
    null_M_cnt = Calloc(Npairs, int);
    null_Mstar_cnt = Calloc(Npairs, int);

    for (pair=0; pair < Npairs; pair++)
    {
        null_M_cnt[pair] = 0;
        null_Mstar_cnt[pair] = 0; 
    }


    /* More misc. storage allocation, etc */
    int *indx;
    indx = Calloc(Nseq, int);

    double *runifVec;
    runifVec = Calloc(Nseq, double);


    /* Initialize R's RNG (Random Number Generator) */
    GetRNGstate();


    /* (3) Repeatedly create "null" data sets, compute statistics based
     *     upon them, and record whether the null stats are more "extreme"
     *     than the stats calculated from the observed data 
     */
    int iter;
    for (iter=0; iter<Nperm; iter++)
    {

        /* reorder each row of 'x' separately */
        for (i=0; i<Npos; i++) 
        {
            /* generate unif(0,1) vals and initialize indx to 0, 1, .., n-1 */
            for (j=0; j < Nseq; j++)
            { 
                runifVec[j] = runif(0,1);
                indx[j] = j;
            }

            /* get indx ordering the random numbers (i.e. random index) */
            indexSort(runifVec, indx, Nseq);

            /* Re-order i-th row of 'x' */
            for (j=0; j < Nseq; j++)
                null_x[i][j] = x[i][indx[j]];
        }


        /* calc statistics on current null data set */
        calcStats(null_x, Nseq, Npos, posI, posJ, Npairs, jntprobIJ, 
                  Nlev, maxLvl, wt, useWt, p, sum_PlogP, null_M,
                  null_Mstar, &anyNAN);

        /* Since the single group analysis uses permutation to estimate p-vals,
         * we should never get NaNs (since we didn't for the original data, 
         * else we wouldn't have made it down here)  But leave in as a check */
        if (anyNAN == TRUE)
        {
            printf("%s", "Encountered unexpected problem w/permuted data:\n");
            printf("%s", "NaN values produced for Mstar in function",
                         " calcStats.  Exiting...\n\n");
            exit(456);
        }

        /* Keep counts of values more extreme than those of observed data */ 
        for (pair=0; pair < Npairs; pair++)
        {
            if (null_M[pair] >= M[pair] - DBL_EPSILON)
                null_M_cnt[pair]++ ;

            if (null_Mstar[pair] >= Mstar[pair] - DBL_EPSILON)
                null_Mstar_cnt[pair]++ ; 
        }

    }  
    /* End of permutation loop */


    /* Since the single group analysis uses permutation to estimate p-vals,
     * we should never get NaNs (since we didn't for the original data, 
     * else we wouldn't have made it to here)  But leave in as a check */
    if (anyNAN == TRUE)
    {
        printf("%s", "Encountered unexpected problem w/permuted data:\n");
        printf("%s", "NaN values produced for Mstar in function",
                     " calcStats.  Exiting...\n\n");
        exit(456);
    }


    /* Calculate raw (unadjusted) pvals for statistics:
     * Pvals computed including the observed value as on of the null  *
     * values.  Note also that the counts of more extreme values are  *
     * one-sided, so we need to double them to get a two-sided pvalue */
    double Nplus1;
    Nplus1 = Nperm + 1; 
    for (pair=0; pair < Npairs; pair++)
    {
        /* Naive computation of raw p-values:
        M_p[pair] = 2*null_M_cnt[pair] / (double) Nperm;
        Mstar_p[pair] = 2*null_Mstar_cnt[pair] / (double) Nperm;
	*/ 

        /* "Better" computation of raw p-values - includes observed count */
        M_p[pair]     = fmin(1, (2*null_M_cnt[pair] + 1)/Nplus1 );
        Mstar_p[pair] = fmin(1, (2*null_Mstar_cnt[pair] + 1)/Nplus1 );
    }


    /* Free all memory allocated with Calloc */ 
    free(x);
    free(p); 
    free(sum_PlogP);
    free(null_M);
    free(null_Mstar);
    free(null_M_cnt);
    free(null_Mstar_cnt);
    free(indx);
    free(runifVec);
    if (useWt != 0) free(wt);

    for (k=0; k<Npos; k++) 
        free(null_x[k]);
    free(null_x); 

    for (k=0; k<maxLvl; k++) 
        free(jntprobIJ[k]);
    free(jntprobIJ);


    /* Copy RNG state back to R */
    PutRNGstate();

    /** We're done! YAY! **/
}  
/* end 'singleGroup' */



/* =========================================================================== */



void compareGroupsPairs(int *xvec, int *yvec, int *ptNseq_x, int *ptNseq_y,
                        int *ptNpos, int *Nlev_x, int *Nlev_y,
                        int *posI, int *posJ, int *ptNpairs,
                        double *xwtvec, double *ywtvec, int *ptuseWt,
                        double *xpvec, double *ypvec, int *ptNboot,
                        int *usableNullSampsCnt,
                        double *M, double *M_p, double *Mstar, double *Mstar_p)
{

    /* define variable types */
    int i, j, k, pair;
    int Nseq_x, Nseq_y, Npos, Npairs, useWt, Nboot;
    long double tmpVal;
     
    Nseq_x = *ptNseq_x;
    Nseq_y = *ptNseq_y;
    Npos = *ptNpos;
    Npairs = *ptNpairs;
    useWt = *ptuseWt;
    Nboot = *ptNboot;


    /*** BEGIN HOUSEKEEPING ***/

    /* All matrices/list from R have to be passed to C as vectors,
     * but for clarity, we'd like to access them as if they are matrices.
     * So we set up the infra-structure to allow us to do that.
     */

    /* Create int matrix 'x' of dim Npos x Nseq_x to access data in 'xvec' */
    /* and int matrix 'y' of dim Npos x Nseq_y to access data in 'yvec' */
    int **x, **y;
    x = Calloc(Npos, int *);
    y = Calloc(Npos, int *);
    for (i = 0; i < Npos; i++) {
         x[i] = & ( xvec[ i * Nseq_x ] );
         y[i] = & ( yvec[ i * Nseq_y ] );
    }


    /* Create 'p_x' and 'p_y', each length Npos, to access marginal prob.s of */
    /* of groups 'x' and 'y' at each position of the seqs*/
    double **p_x, **p_y;
    int loc_x, loc_y;

    p_x = Calloc(Npos, double *);
    p_y = Calloc(Npos, double *);

    loc_x = 0;
    loc_y = 0;

    for (i = 0; i < Npos; i++)
    {
         p_x[i] = & ( xpvec[loc_x] );
         loc_x += Nlev_x[i];

         p_y[i] = & ( ypvec[loc_y] );
         loc_y += Nlev_y[i];
    }


    /* If useWt != 0, create 'Nseq_x' x 'Nseq_x' matrix 'wt_x' to access xwtvec */
    /* and 'Nseq_y' x Nseq_y matrix wt_y to access ywtvec */
    double **wt_x, **wt_y;
    if (useWt != 0)
    {
        wt_x = Calloc(Nseq_x, double *);
        wt_y = Calloc(Nseq_y, double *);

        for (j = 0; j < Nseq_x; j++)
             wt_x[j] = & ( xwtvec[ j * Nseq_x ] );

        for (j = 0; j < Nseq_y; j++)
             wt_y[j] = & ( ywtvec[ j * Nseq_y ] );
    }



    /* Next, we set up a matrix to store jnt freq. results for a pair of pos.,
     * then pre-calculate some quantities needed for statistic Mstar.
     */

    /* Find the maximum number of levels across all positions and both x and y */
    int maxLvl;
    maxLvl = 0;
    for (i = 0; i < Npos; i++) 
    {
        if (Nlev_x[i] > maxLvl)
            maxLvl = Nlev_x[i];
        if (Nlev_y[i] > maxLvl)
            maxLvl = Nlev_y[i];
    } 


    /* Create a matrix of dimension (maxLvl x maxLvl) to be used by
     * function 'calcStats' for storing jnt freqs of pos i and j  
     */
    double **jntprobIJ;
    jntprobIJ = Calloc(maxLvl, double *);
    for (k = 0; k < maxLvl; k++)
        jntprobIJ[k] = Calloc(maxLvl, double);


    /* Pre-calculate terms needed for statistic Mstar */
    double *sum_PlogP_x,  *sum_PlogP_y;
    sum_PlogP_x = Calloc(Npos, double);
    sum_PlogP_y = Calloc(Npos, double);

    for (i = 0; i < Npos; i++)
    {
       tmpVal = 0;
       for (k = 0; k < Nlev_x[i]; k++)
           tmpVal += p_x[i][k] * log( p_x[i][k] );

       sum_PlogP_x[i] = tmpVal;


       tmpVal = 0;
       for (k = 0; k < Nlev_y[i]; k++)
           tmpVal += p_y[i][k] * log( p_y[i][k] );

       sum_PlogP_y[i] = tmpVal;
    }

    /* Storage for individual group statistics */
    double *M_x, *M_y, *Mstar_x, *Mstar_y;
    M_x = Calloc(Npairs, double);
    M_y = Calloc(Npairs, double);
    Mstar_x = Calloc(Npairs, double);
    Mstar_y = Calloc(Npairs, double);

    /* Storage for sign of statistics */
    int *sign_M, *sign_Mstar;
    sign_M = Calloc(Npairs, int);
    sign_Mstar = Calloc(Npairs, int);

    /*** HOUSEKEEPING DONE ***/



 
    /*** NOW LETS DO SOME WORK! ***/

    /* define/initialize indicator of NAN occurrance in calcStats() */
    int anyNAN;
    anyNAN = FALSE;

    /* (1) calculate statistics for set of original data 'x' and 'y' */
    calcStats(x, Nseq_x, Npos, posI, posJ, Npairs, jntprobIJ,
              Nlev_x, maxLvl, wt_x, useWt, p_x, sum_PlogP_x,
              M_x, Mstar_x, &anyNAN);

    calcStats(y, Nseq_y, Npos, posI, posJ, Npairs, jntprobIJ, 
              Nlev_y, maxLvl, wt_y, useWt, p_y, sum_PlogP_y,
              M_y, Mstar_y, &anyNAN);

    
    /* This code should never run, as original data was screened to remove *
     * remove positions with no variation in them. But just in case ...    */
    if (anyNAN == TRUE)
    {
        printf("%s", "Encountered unexpected problem w/original data:\n");
        printf("%s", "NaN values produced for Mstar in function",
                     " calcStats.  Exiting...\n\n");
        exit(456);
    }


    /* Now calculate the difference statistics: the difference between *
     * the values of the two groups' statistics                        */
    for (pair=0; pair < Npairs; pair++)
    {
        M[pair] = M_x[pair] - M_y[pair];
        Mstar[pair] = Mstar_x[pair] - Mstar_y[pair];

        sign_M[pair] = Sign(M[pair]);
        sign_Mstar[pair] = Sign(Mstar[pair]);
    }



    /* (2) Define/allocate/initialize storage for null data, null statistics,
     *     and for counts of null statistics more extreme than observed ones
     */
    int **x_boot, **y_boot;
    x_boot = Calloc(Npos, int *);
    y_boot = Calloc(Npos, int *);
    for (i=0; i<Npos; i++)
     {
        x_boot[i] = Calloc(Nseq_x, int);
        y_boot[i] = Calloc(Nseq_y, int);
     }

    /* More misc. storage allocation, etc */
    int *x_boot_indx, *y_boot_indx;
    x_boot_indx = Calloc(Nseq_x, int);
    y_boot_indx = Calloc(Nseq_y, int);

    double *boot_M, *boot_Mstar, *boot_M_x, *boot_M_y,
           *boot_Mstar_x, *boot_Mstar_y;
    boot_M = Calloc(Npairs, double);
    boot_Mstar = Calloc(Npairs, double);
    boot_M_x = Calloc(Npairs, double);
    boot_M_y = Calloc(Npairs, double);
    boot_Mstar_x = Calloc(Npairs, double);
    boot_Mstar_y = Calloc(Npairs, double);

    /* counts of number of NaN values produced - these are not used */
    int *boot_NANcnt; 
    boot_NANcnt = Calloc(Npairs, int);

    /* counts of bootstrapped stats "on the other side of zero" from stats*/
    int *boot_M_cnt, *boot_Mstar_cnt;
    boot_M_cnt = Calloc(Npairs, int);
    boot_Mstar_cnt = Calloc(Npairs, int);

    /* initialize count vectors to zero */ 
    for (pair=0; pair < Npairs; pair++)
    {
        boot_NANcnt[pair] = 0;
        boot_M_cnt[pair] = 0;
        boot_Mstar_cnt[pair] = 0;
    }


    /* Initialize R's RNG (Random Number Generator) */
    GetRNGstate();


    /* (3) Repeatedly bootstrap the data sets, compute statistics based
     *     upon them, and record whether the stats ... 
     */
    int iter;
    for (iter=0; iter<Nboot; iter++)
    {
        /* get indices of seqs that comprise the current bootstrap sample */
        /* Values are in the set {0, 1,..., Nseq -1 } */ 
        for (j = 0; j < Nseq_x; j++) 
           x_boot_indx[j] = floor( Nseq_x * runif(0,1) );

        for (j = 0; j < Nseq_y; j++) 
           y_boot_indx[j] = floor( Nseq_y * runif(0,1) );


        /* create bootstrap data sets*/
        for (i=0; i<Npos; i++)
        {
            for(j=0; j<Nseq_x; j++) 
                x_boot[i][j] = x[i][ x_boot_indx[j] ];

            for(j=0; j<Nseq_y; j++) 
                y_boot[i][j] = y[i][ y_boot_indx[j] ]; 
        } 


        /* Calculate vectors p for 'x_boot' and 'y_boot' */
        for (i = 0; i < Npos; i++)
        {
            freq1D(x[i], Nlev_x[i], Nseq_x, p_x[i]);
            freq1D(y[i], Nlev_y[i], Nseq_y, p_y[i]);
        }

        /* Pre-calculate terms needed for statistic Mstar */
        /* Note - we are re-using the same objects used for original data values */
        double p_ik;
        for (i = 0; i < Npos; i++)
        {
           sum_PlogP_x[i] = 0;
           for (k = 0; k < Nlev_x[i]; k++) 
           {
               p_ik = p_x[i][k]; 
               if ( p_ik  > 0 )
                   sum_PlogP_x[i] += p_ik * log( p_ik );
           }

           sum_PlogP_y[i] = 0;
           for (k = 0; k < Nlev_y[i]; k++)
           {
               p_ik = p_y[i][k]; 
               if ( p_ik  > 0 )
                   sum_PlogP_y[i] += p_ik * log( p_ik );
           }
        }


        /* calc stats for each set of bootstrapped data 'x_boot' and 'y_boot' */
        calcStats(x_boot, Nseq_x, Npos, posI, posJ, Npairs, jntprobIJ,
                  Nlev_x, maxLvl, wt_x, useWt, p_x, sum_PlogP_x,
                  boot_M_x, boot_Mstar_x, &anyNAN);

        calcStats(y_boot, Nseq_y, Npos, posI, posJ, Npairs, jntprobIJ,
                  Nlev_y, maxLvl, wt_y, useWt, p_y, sum_PlogP_y,
                  boot_M_y, boot_Mstar_y, &anyNAN);


        /* calc difference in stats between the two groups */
        double bM, bMs;
        for (pair=0; pair < Npairs; pair++)
        {
            bM  = boot_M_x[pair] - boot_M_y[pair];
            bMs = boot_Mstar_x[pair] - boot_Mstar_y[pair];

            boot_M[pair] = bM;
            boot_Mstar[pair] = bMs;

            /* count bootstrapped stats on "other side of zero" from stats */
            if (anyNAN == FALSE || !isnan(bMs))
            {
                if ( Sign(bM) != sign_M[pair] ||  fabs(bM) < DBL_EPSILON )
                    boot_M_cnt[pair]++ ;

                if ( Sign(bMs) != sign_Mstar[pair] || fabs(bMs)< DBL_EPSILON )
                    boot_Mstar_cnt[pair]++ ;
            } else
            {
                boot_NANcnt[ pair ]++ ;
            }
        }

     	/* reset NAN indicator for next bootstrap sample */
        if (anyNAN == TRUE)
            anyNAN = FALSE;

    }
    /* End bootstrap loop */


    /*   Calculate raw (unadjusted) pvals for statistics. */
    double nonNANcnt_plus1;
    for (pair = 0; pair < Npairs; pair++)
    {
        usableNullSampsCnt[pair] =  Nboot - boot_NANcnt[pair];

        /* Pvals computed including the observed value as on of the null  *
         * values.  Note also that the counts of more extreme values are  *
         * one-sided, so we need to double them to get a two-sided pvalue */
        nonNANcnt_plus1 = usableNullSampsCnt[pair] + 1;

        M_p[pair]     = fmin(1, (2*boot_M_cnt[pair] + 1)/nonNANcnt_plus1);
        Mstar_p[pair] = fmin(1, (2*boot_Mstar_cnt[pair] + 1) / nonNANcnt_plus1);
    }


    /* Free all memory allocated with Calloc */
    free(x);
    free(y);
    free(p_x); 
    free(p_y); 
    free(sum_PlogP_x);
    free(sum_PlogP_y);
    free(x_boot_indx);
    free(y_boot_indx);
    free(M_x);
    free(M_y);
    free(Mstar_x);
    free(Mstar_y);
    free(sign_M);
    free(sign_Mstar);

    free(boot_M);
    free(boot_Mstar);

    free(boot_M_x);
    free(boot_Mstar_x);
    free(boot_M_y);
    free(boot_Mstar_y);
    free(boot_M_cnt);
    free(boot_Mstar_cnt);
    free(boot_NANcnt); 

    if (useWt != 0) 
    {
        free(wt_x);
        free(wt_y);
    }

    for (k=0; k<Npos; k++) 
    {
        free(x_boot[k]);
        free(y_boot[k]);
    }
    free(x_boot); 
    free(y_boot); 

    for (k=0; k<maxLvl; k++) 
        free(jntprobIJ[k]);
    free(jntprobIJ);

    /* Copy RNG state back to R */
    PutRNGstate();

}
/* end 'compareGroups' */

