## Functions defined in this file:  CovDoubles, singleGrpPairs, compareGrpsPairs

## Include code for functions: prepare.posPairs, update.posPairs
src.dir <- "/.../..."
source( file.path(src.dir, "posPairsUtils.R") )

## Uses C code from file Doubles.c, compiled using "R CMD SHLIB Doubles.c"
dyn.load( file.path(src.dir, "Doubles.so") )


covDoubles <- function(grp1, grp2=NULL, wt1=NULL, wt2=NULL, Niter=20000,

                   ## min number of non-consensus (nc) AAs that must be present
                   ## among the seqs at a pos for that position to be used
                   grp1.min.nc = 3,
                   grp2.min.nc = 3,

                   ## posPairs should be a two-column numeric matrix giving
                   ## the pairs of positions: (col. 1, col. 2) to be tested.
                   ## The default is to consider all unique pairs.
                   posPairs = NULL,

                   # code used to represent gaps,
                   gapCode = "-",

                   # vector of all acceptable codes,
                   validCodes = c("A","C","D","E","F","G","H","I","K",
                                  "L","M","N","P","Q","R","S","T","V",
                                  "W","Y", gapCode), 

                   # action to take if codes other than 'validCodes' are found
                   otherCodes = c("recode","exit")
                   ) 
{

    ###### Begin argument checking:   ##########################################
    
    ##  - 'grp1' must be a matrix 
    if (!is.matrix(grp1)) 
        stop("'grp1' must be a matrix")

    ##  - if not NULL, 'grp2' must be a matrix with same number of rows as grp1
    if (!is.null(grp2)) 
    {
        if (!is.matrix(grp2)) 
            stop("'grp2' must be a matrix\n")
        else if (nrow(grp1) != nrow(grp2))
            stop("Matrices 'grp1' and 'grp2' must have same number of rows\n")
    }

    ##  - Niter must be > 0
    if ( !(Niter > 0) )
        stop("Arg. 'Niter' must be an integer > 0\n\n")

    ##  - both grp1.min.nc and grp2.min.nc must be >= 1
    if ( min(grp1.min.nc, grp2.min.nc) < 1 )
        stop("Arg.s 'grp1.min.nc' and 'grp2.min.nc' must both be >= 1\n\n")

    ##  - 'posPairs' should be a numeric matrix with 2 columns
    if (!is.null(posPairs)) 
    {
        if (!is.numeric(posPairs)) stop("Arg. posPairs must be numeric\n") 
        if (!is.matrix(posPairs)) stop("Arg. posPairs must be a 2 column matrix\n") 
        if (ncol(posPairs)!=2) stop("Arg. posPairs must be a 2 column matrix\n") 
    }

    ##  - 'gapCode' should be a character string of length 1
    if ( !(is.character(gapCode) && length(gapCode)==1 && nchar(gapCode)==1) ) 
        stop("\ngapCode is mis-specified - must be a char. string of length 1")

    ##  - 'validCodes' should be a char vector, w/each string of length 1
    if ( !(is.character(validCodes) && all(nchar(validCodes)==1)) )
        stop("'validCodes' is mis-specified\n")

    ##  - 'otherCodes' must be one of the strings given in the function defn
    ##    The default value is the first appearing in the vector of choices
    otherCodes <- match.arg(otherCodes)


    ##  - 'wt1' should be numeric and have dimension ncol(grp1)x ncol(grp1)
    if (!is.null(wt1)) 
        if ( !(is.numeric(wt1) && all( dim(wt1) == ncol(grp1) ) ))
            stop("'wt1' must be numeric symmetric matrix, w/dims = ncol(grp1)\n")

    ##  - 'wt2' should be numeric and have dimension ncol(grp2)x ncol(grp2)
    if (!is.null(wt2)) 
        if ( !(is.numeric(wt2) && all( dim(wt2) == ncol(grp2) ) ))
            stop("'wt2' must be numeric symmetric matrix, w/dims = ncol(grp2)\n")

    ##  - If grp2 given, then 'wt1' and 'wt2' should both be null or both given
    if (!is.null(grp2)) 
        if ( xor(is.null(wt1), is.null(wt2)) )
            stop("'wt1' and 'wt2' must both be null or both be specified\n")

    ###### End argument checking:   ############################################



    ## Define 'Npos': the number of positions in the original
    ## input sequences (i.e. rows in grp1)
    Npos <- nrow(grp1) 



    ####### Begin group pre-processing:  #######################################
    
    ## Check values of characters in grp1/grp2 against values in 'validCodes'.
    ## If codes not in 'validCodes' are found, either recode or exit depending 
    ## on the value of argument 'otherCodes'

    w.notValid <- which(!grp1 %in% validCodes)
    if (length(w.notValid) > 0) 
    {
        if ('otherCodes' == "exit") 
            stop("Invalid codes present in 'grp1'. Exiting...\n")
        else
            grp1[w.notValid] <- gapCode 
    }

    if (!is.null(grp2))
    { 
        w.notValid <- which(!grp2 %in% validCodes)

        if (length(w.notValid) > 0) 
        {
            if ('otherCodes' == "exit") 
                stop("Invalid codes present in 'grp2'. Exiting...\n")
            else
                grp2[w.notValid] <- gapCode 
        }
    }


    ## posPairs processing
    ## If posPairs was not user-defined, then create it, including all pairs
    ## If posPairs was defined, check it and (if nec.) subset and order
    if (is.null(posPairs))
    {
        user.defined.posPairs <- FALSE

        ## Create our own matrix 'posPairs' containing all pairs
        posPairs <- prepare.posPairs("create", posPairs, Npos)$posPairs

    } else 
    {
        user.defined.posPairs <- TRUE

        ## Examine user provided matrix 'posPairs' and modify/order if needed
        res <- prepare.posPairs("process", posPairs, Npos)

        ## over-write with new 'posPairs' (may or may-not have changed)
        posPairs <- res$posPairs


        if (res$anyToDrop == TRUE)
        {
            ## Remove positions from the data that weren't found in posPairs
            grp1 <- grp1[-res$dropPos, ]
            if (!is.null(grp2)) 
                grp2 <- grp2[-res$dropPos, ]

            ## Re-code the remaining positions to 1,2,3... after excluding 
            ## 'dropPos', and return the original positions (for later remapping)
            update <- update.posPairs("RecodeOnly", res$dropPos, posPairs, Npos)

            ## over-write with new, re-coded 'posPairs' 
            posPairs <- update$posPairs

            ## retain indices of positions within the original data
            origPos <- update$origPos

            ## set flag indicating that we've recoded values
            ReCoded.posPairs <- TRUE

            rm(update)
        }  else  
            ReCoded.posPairs <- FALSE

        rm(res)
    }
    
    ## Get consensus seq(s) for positions being used in groups
    consensus.fun <- function(x) { tbl <- table(x)
                                   names(tbl)[ which.max(tbl) ] }
    ConsensusSeq1 <- apply(grp1, 1, consensus.fun)
    if (!is.null(grp2)) 
        ConsensusSeq2 <- apply(grp2, 1, consensus.fun)


    ## Convert grp1/grp2 into integer matrix using a row-specific encoding.
    ## Each row gets encoded separately, so same codes represent different
    ## amino acids in diff. rows
    
    encod <- function(x) as.integer(factor(x))

    g1 <- t(apply(grp1, 1, encod))
    if (!is.null(grp2)) 
        g2 <- t(apply(grp2, 1, encod))




    ## Now run the analysis on the group or groups

    if (is.null(grp2)) 
    {
        # - Analysis of covariation of pairs of positions in grp1:
        out <- singleGrpPairs(g1, wt1, posPairs, grp1.min.nc, Niter)

    } else
    {
        # - comparison of covariation of pairs of positions in Grp1 and Grp2
        out <- compareGrpsPairs(g1, g2, wt1, wt2, grp1.min.nc, grp2.min.nc,
                                posPairs, Niter)
    }



    ## ---------------------------------------- ##
    ## Make some additions to the output object ##
    ## ---------------------------------------- ##

    ## Add info regarding the function invocation 
    out <- c(list(Call = list(call = match.call(),
                         Defaults = formals())), out)


    ## Add info on whether user specified position pairs
    out$Analysis$posPairsSpecified <- user.defined.posPairs 


    ## Add consensus seq info to output
    if (is.null(grp2))
    {
        w.pos <- out$Data$Analyzed$posList 
        out$Data$Analyzed$ConsensusSeq <- ConsensusSeq1[w.pos] 
    } else
    {
        w.pos <- out$Data$Analyzed$posList 
        out$Data$Analyzed$ConsensusSeq <- list(Grp1=ConsensusSeq1[w.pos],
                                               Grp2=ConsensusSeq2[w.pos]) 
    }

    ## get adjusted pvalues
    out$Results <- transform(out$Results, 
                             Mp.bon = p.adjust(Mp.raw, method="bonf"),
                             Mp.fdr = p.adjust(Mp.raw, method="fdr"),
                             Mstarp.bon = p.adjust(Mstarp.raw, method="bonf"),
                             Mstarp.fdr = p.adjust(Mstarp.raw, method="fdr"))

    ## add info on numbers of pairs significant by various criteria
    out$Signif <- list(Mp05.fdr = sum(out$Results$Mp.fdr < .05),
                       Mp10.fdr = sum(out$Results$Mp.fdr < .10),
                       Mp05.bon = sum(out$Results$Mp.bon < .05),
                       Mp10.bon = sum(out$Results$Mp.bon < .10),
                       Mstarp05.fdr = sum(out$Results$Mstarp.fdr < .05),
                       Mstarp10.fdr = sum(out$Results$Mstarp.fdr < .10),
                       Mstarp05.bon = sum(out$Results$Mstarp.bon < .05),
                       Mstarp10.bon = sum(out$Results$Mstarp.bon < .10))

  
    ## *If* posPairs was specified, and some row dropping was done
    ## earlier, then convert back to original data positions
    if (user.defined.posPairs && ReCoded.posPairs)
    { 
        out$Results$posI <- origPos[ out$Results$posI ]
        out$Results$posJ <- origPos[ out$Results$posJ ]
        out$Data$Analyzed$posList <- origPos[ out$Data$Analyzed$posList ] 
        out$Data$Excluded$posList <- origPos[ out$Data$Excluded$posList ] 
    }

    return(out)
}



singleGrpPairs <- function(g, wt, posPairs, min.nc, Nperm)
{

    ## Number of sequences in data
    Nseq <- ncol(g)
    
    ## Calculate the marginal counts for each row.
    ## This will produce a list as pos. will have diff length marginals
    marg.cnts <- apply(g, 1, tabulate)

    ## N.c  = number of occurrances of "consensus" (most common) value
    ## N.nc = number of occurrances of non-"consensus" values
    N.c <- unlist(lapply(marg.cnts, max))
    N.nc <- Nseq - N.c

    ## Figure out which (if any) rows should be excluded from the analysis.
    ## We exclude rows that are too homogeneous (i.e. for which most seqs
    ## have the same value).
    w.excl <- which(N.nc < min.nc)
    Nexcl <- length(w.excl)

    ## Store original number of position Pairs, provide info in output obj.
    orig.Npairs <- nrow(posPairs)

    ## if any positions don't meet criteria, they are excluded, then we
    ## do some housekeeping
    if (Nexcl > 0) 
    {
        cat("\n\nExcluded ", Nexcl, " positions from analysis due to ",
            "insufficient variability.\n", nrow(g) - Nexcl,
            " positions remain.\n\n", sep="")

        ## Now that we've excluded do a last check to be sure we have
        ## enough positions remaining to run analysis.
        if (nrow(g) - Nexcl < 2)
            stop("\nMust have at least 2 positions to run analysis;  ",
                 "Exiting...\n\n")


        ## update posPairs by removing pairs involving excluded positions
        ## and then recoding positions back to 1,2,3,...
        update <- update.posPairs("dropThenRecode", w.excl, posPairs, Npos=nrow(g))

        ## updated posPairs
        posPairs <- update$posPairs

        ## retain index to original position values
        origPos <- update$origPos

        excluded.pos <- update$excluded

        g <- g[ -excluded.pos, ]
        marg.cnts <- marg.cnts[ -excluded.pos ]

        rm(update) 
    } 

    ## Number of unique pairs of positions (i,j) to be examined
    Npairs <- nrow(posPairs)

    ## Number of positions in data set that will be used
    Npos <- nrow(g)

    ## Marginal probability for the levels present at each position.
    ## calculated from marginal counts 
    marg.p <- lapply(marg.cnts, function(cnts, N) cnts/N , N=Nseq)

    ## Nlev gives the number of "levels" (unique values) for each pos.
    Nlev <- unlist(lapply(marg.p, length))


    rm(marg.cnts, N.c, N.nc, w.excl) 
    gc()
    
    ## ********************************************************************** ##  
    ## Note regarding matrices being passed to C code:
    ##   - Since all objects must be passed as vectors to C, any objects we
    ##     wish to reconstruct into matrices inside the C code must be
    ##     transposed before vectorization via as.integer, as.double, etc.
    ##     This is because the vectorization is done "by column" and the 
    ##     resulting vector can't be reproduced as a matrix of the same dim. 
    ##     of the original R object.
    ## ********************************************************************** ##  

    t0 <- proc.time()[3]
    out <- .C("singleGroupPairs", 
              g = as.integer(t(g)), 
              Nseq = as.integer(Nseq),
              Npos = as.integer(Npos),
              Nlev = as.integer(Nlev),
              posI = as.integer(posPairs[,"i"]),
              posJ = as.integer(posPairs[,"j"]),
              Npairs = as.integer(Npairs),

              wt = as.double(wt),
              use.wt = as.integer(length(wt)),  
              ## use.wt==0 when wt==NULL, else use.wt>=1 

              p = as.double(unlist(marg.p)), 
              Nperm = as.integer(Nperm), 

              ## values below are calculated in program
              M = as.double(numeric(Npairs)),
              Mp.raw = as.double(numeric(Npairs)),
              Mstar = as.double(numeric(Npairs)),
              Mstarp.raw = as.double(numeric(Npairs))
             ) 

    t1 <- proc.time()[3]
    cat("Time in minutes for C code =", round((t1-t0)/60,1), "\n\n")

    ## construct output object to return to function 'covDoubles'
    out <- list(Analysis = list(
                    Type = "Single group: Covariation of Pairs of positions",
                    weightsUsed = isTRUE(out$use.wt > 0),
                    posPairsSpecified = NULL,   ## value provided in covDoubles
                    Min.NonCons.needed = min.nc,
                    Niter = out$Nperm
                    ),
                Data = list(
                    Analyzed = list(
                        Nseq = out$Nseq,
                        Npos = out$Npos,
                        Npairs = out$Npairs,
                        posList = if (Nexcl>0) origPos else (1:out$Npos)
                        ),
                    Excluded = list(
                        Npos = Nexcl,
                        Npairs = orig.Npairs - out$Npairs,
                        posList = if (Nexcl>0) excluded.pos else NULL
                        )
                    ),
                Results = as.data.frame(
                    out[c("posI", "posJ", "M", "Mp.raw", "Mstar", "Mstarp.raw")]
                    )
               )

    ## Translate positions back to original coordinates (if necessary)
    if (Nexcl > 0)
    {
        out$Results$posI <- origPos[ out$Results$posI ] 
        out$Results$posJ <- origPos[ out$Results$posJ ] 
    }
   
    return(out)
}



compareGrpsPairs <- function(g1, g2, wt1, wt2, g1.min.nc, g2.min.nc, 
                             posPairs, Nboot)
{
    ## Number of sequences in data
    N1seq <- ncol(g1)
    N2seq <- ncol(g2)

    ## Calculate the marginal counts for each row (Note: output is a list)
    g1.marg.cnts <- apply(g1, 1, tabulate)
    g2.marg.cnts <- apply(g2, 1, tabulate)

    ## N.c  = number of occurrances of "consensus" (most common) value
    ## N.nc = number of occurrances of non-"consensus" values
    N1.c <- unlist(lapply(g1.marg.cnts, max))
    N1.nc <- N1seq - N1.c
    N2.c <- unlist(lapply(g2.marg.cnts, max))
    N2.nc <- N2seq - N2.c

    ## Figure out which (if any) rows should be excluded from the analysis.
    ## We exclude rows that are too homogeneous (i.e. for which most seqs
    ## have the same value).
    w.excl.g1 <- which(N1.nc < g1.min.nc)
    w.excl.g2 <- which(N2.nc < g2.min.nc)

    ## Exclude positions found on *either* of the groups' exclude lists
    w.excl <- unique( c(w.excl.g1, w.excl.g2) )
    Nexcl <- length(w.excl)

    ## Store original number of position Pairs, provide info in output obj.
    orig.Npairs <- nrow(posPairs)

    ## If there are positions to be excluded we do so, then clean up the mess
    if (Nexcl > 0)
    {
        cat("\n\nExcluded ", Nexcl, " positions from analysis due to ",
            "insufficient variability.\n", nrow(g1) - Nexcl,
            " positions remain.\n\n", sep="")

        ## Now that we've excluded do a last check to be sure we have
        ## enough positions remaining to run analysis.
        if (nrow(g1) - Nexcl < 2)
            stop("\nMust have at least 2 positions to run analysis;  ",
                 "Exiting...\n\n")


        ## update posPairs by removing pairs involving excluded positions
        ## and then recoding positions back to 1,2,3,...
        update <- update.posPairs("dropThenRecode", w.excl, posPairs, nrow(g1))

        ## updated posPairs
        posPairs <- update$posPairs

        ## retain index to original position values
        origPos <- update$origPos

        excluded.pos <- update$excluded

        ## remove excluded positions from data 
        g1 <- g1[ -excluded.pos, ]
        g2 <- g2[ -excluded.pos, ]
        g1.marg.cnts <- g1.marg.cnts[ -excluded.pos ]
        g2.marg.cnts <- g2.marg.cnts[ -excluded.pos ]

        rm(update)
    }

    ## Number of unique pairs of positions (i,j) to be examined
    Npairs <- nrow(posPairs)

    ## Number of positions in data set that will be used
    Npos <- nrow(g1)

    ## Compute marginal probability from marginal counts
    g1.marg.p <- lapply(g1.marg.cnts, function(cnts, N) cnts/N , N=N1seq)
    g2.marg.p <- lapply(g2.marg.cnts, function(cnts, N) cnts/N , N=N2seq)

    ## Nlev gives the number of "levels" (unique values) for each pos.
    N1lev <- unlist(lapply(g1.marg.p, length))
    N2lev <- unlist(lapply(g2.marg.p, length))

    rm(g1.marg.cnts, g2.marg.cnts, N1.c, N1.nc, N2.c, N2.nc,
       w.excl.g1, w.excl.g2, w.excl)
    gc()

    t0 <- proc.time()[3]
    out <- .C("compareGroupsPairs", 
               g1 = as.integer(t(g1)), 
               g2 = as.integer(t(g2)), 
               N1seq = as.integer(N1seq),
               N2seq = as.integer(N2seq),
               Npos = as.integer(Npos),
               N1lev = as.integer(N1lev),
               N2lev = as.integer(N2lev),
               posI = as.integer(posPairs[,"i"]),
               posJ = as.integer(posPairs[,"j"]),
               Npairs = as.integer(Npairs),

               wt1 = as.double(wt1),
               wt2 = as.double(wt2),
               use.wt = as.integer( length(wt1)>0 ),  
               ## use.wt==0 when wt1==NULL and wt2==NULL, else use.wt>=1 

               p1 = as.double(unlist(g1.marg.p)), 
               p2 = as.double(unlist(g2.marg.p)), 
               Nboot = as.integer(Nboot), 

               ## values below are calculated in program
               N.samps = integer(Npairs), 
               M = as.double(numeric(Npairs)),
               Mp.raw = as.double(numeric(Npairs)),
               Mstar = as.double(numeric(Npairs)),
               Mstarp.raw = as.double(numeric(Npairs))
              ) 

    t1 <- proc.time()[3]
    cat("Time in minutes for C code =", round((t1-t0)/60,1), "\n\n")


    ## construct output object to return to function 'covDoubles'
    out <- list(Analysis = list(
                    Type = "Two group comparison: Diff. in Covar. of pairs of positions",
                    weightsUsed = isTRUE(out$use.wt > 0),
                    posPairsSpecified = NULL,   ## value provided in covDoubles
                    Min.NonCons.needed = c(Grp1 = g1.min.nc, Grp2=g2.min.nc),
                    Niter = out$Nboot
                    ),
                Data = list(
                    Analyzed = list(
                        Nseq = c(Grp1=out$N1seq, Grp2=out$N2seq),
                        Npos = out$Npos,
                        Npairs = out$Npairs,
                        posList = if (Nexcl>0) origPos else (1:out$Npos)
                        ),
                    Excluded = list(
                        Npos = Nexcl,
                        Npairs = orig.Npairs - out$Npairs,
                        posList = if (Nexcl>0) excluded.pos else NULL
                        )
                    ),
                Results = as.data.frame(
                    out[c("posI", "posJ", "N.samps", "M", "Mp.raw",
                          "Mstar", "Mstarp.raw")]
                    )
               )


    ## Translate positions back to original coordinates (if necessary)
    if (Nexcl > 0)
    {
        out$Results$posI <- origPos[ out$Results$posI ]
        out$Results$posJ <- origPos[ out$Results$posJ ]
    }

    return(out)
}

# q(save="no")
