## Functions in this file:  CovTriples, singleGrpTriples, compareGrpsTriples

## Include code for functions: create.posTriples, update.posTriples
src.dir <- "/.../..."
source( file.path(src.dir, "posTriplesUtils.R") )

## Uses C code from file Triples.c, compiled using "R CMD SHLIB Triples.c"
dyn.load( file.path(src.dir, "Triples.so") )


covTriples <- function(grp1, grp2=NULL, usePairs=NULL,
                       Niter=20000,
                       grp1.min.nc = 3, grp2.min.nc = 3,
                       gapCode = "-",
                       validCodes = c("A","C","D","E","F","G","H","I","K",
                                      "L","M","N","P","Q","R","S","T","V",
                                      "W","Y", gapCode), 
                       otherCodes = c("recode","exit")
                      )
{
    ## Info on arguments:
    ##
    ## grp1 / grp2
    ## -----------
    ##   Matrices containing one column per sequence and one row per position.
    ##   The matrices may be character or numeric (integer codes), but not yet
    ##   set up for factors (though not too hard to do)
    ##
    ## usePairs 
    ## --------
    ##   'usePairs' must be a two-column numeric matrix giving the pairs of
    ##   positions: (i, j) i<j to be used as the basis for all triplets to be
    ##   tested.  Triplets are formed by taking pairs (i,j) from 'usePairs' and
    ##   joining them with all possible positions k (k!=i and k!=j) to form
    ##   triplets (i,j,k).  Redundant triples ARE REMOVED; this is important
    ##   for correct pvalue adjustment and to help reduce computation time.
    ##   The first occurrance of a triple is one that is retained.
    ##   !!! THIS ARGUMENT *MUST* BE SPECIFIED !!!
    ##  
    ## Niter
    ## -----
    ##   Number of permutations or bootstrap samples to use to estimate
    ##   the un-adjusted p-values for the statistics.
    ##
    ## grpX.min.nc
    ## -----------
    ##   The minimum number of non-consensus (nc) AAs that must be present
    ##   among the sequences at a position, for that position to be included
    ##   in the analysis
    ##
    ## gap.code
    ## --------
    ##   Indicates what 'code' (single character) is used in data sets grp1
    ##   and grp2 to represent a gap.
    ##   
    ## validCodes
    ## ----------
    ##   Vector listing all allowable (i.e. 'valid') codes
    ##   
    ## otherCodes
    ## ----------
    ##   Char. string indicating the action to take if any "other codes" (ones
    ##   not in 'validCodes') are found in the data.  Possible values are:
    ##   'recode' (i.e. replace value with 'gap.code') and 'exit' 


    ###### Begin argument checking:   ##########################################
    
    ##  - 'grp1' must be a matrix
    if (!is.matrix(grp1))
        stop("'grp1' must be a matrix\n\n")

    ##  - if not NULL, 'grp2' must be a matrix with same number of rows as grp1
    if (!is.null(grp2)) 
    {
        if (!is.matrix(grp2)) 
            stop("'grp2' must be a matrix\n\n")
        else if (nrow(grp1) != nrow(grp2))
            stop("Matrices 'grp1' and 'grp2' must have same number of rows\n\n")
    }
 
    ##  - Niter must be > 0
    if ( !(Niter > 0) )
        stop("Arg. 'Niter' must be an integer > 0\n\n")

    ##  - both grp1.min.nc and grp2.min.nc must be >= 1
    if ( min(grp1.min.nc, grp2.min.nc) < 1 )
        stop("Arg.s 'grp1.min.nc' and 'grp2.min.nc' must both be >= 1\n\n") 

    ##  - 'usePairs' should be a numeric matrix with 2 columns
    if (is.null(usePairs)) 
        stop("Argument 'usePairs' must be specified\n\n")
    else 
    {
        if (!is.numeric(usePairs))
            stop("Arg. usePairs must be a numeric matrix\n\n") 
        if (!is.matrix(usePairs) || ncol(usePairs)!=2)
            stop("Arg. usePairs must be a 2 column matrix\n\n") 
    }

    ##  - 'gapCode' should be a character string of length 1
    if ( !(is.character(gapCode) && length(gapCode)==1 && nchar(gapCode)==1) ) 
        stop("\ngapCode mis-specified - must be char. string of length 1\n\n")

    ##  - 'validCodes' should be a char vector, w/each string of length 1
    if ( !(is.character(validCodes) && all(nchar(validCodes)==1)) )
        stop("'validCodes' is mis-specified\n\n")

    ##  - 'otherCodes' must be one of the strings given in the function defn
    ##    The default value is the first appearing in the vector of choices
    otherCodes <- match.arg(otherCodes)

    ###### End argument checking:   ############################################



    ## Define logical var. 'twoGroups' indicating whether two groups
    ## were specified (i.e. whether grp2 is non-NULL)
    twoGroups <- if (!is.null(grp2)) TRUE else FALSE


    ## Define 'Npos': the number of positions in the original
    ## input sequences (i.e. rows in grp1)
    Npos <- nrow(grp1) 


    ####### Begin group pre-processing:  #######################################
    
    ## Check values of characters in grp1/grp2 against values in 'validCodes'.
    ## If codes not in 'validCodes' are found, either recode or exit depending 
    ## on the value of argument 'otherCodes'

    w.notValid <- which(!grp1 %in% validCodes)
    if (length(w.notValid) > 0) 
    {
        if ('otherCodes' == "exit") 
            stop("Invalid codes present in 'grp1'. Exiting...\n\n")
        else
            grp1[w.notValid] <- gapCode 
    }

    if (twoGroups)
    { 
        w.notValid <- which(!grp2 %in% validCodes)

        if (length(w.notValid) > 0) 
        {
            if ('otherCodes' == "exit") 
                stop("Invalid codes present in 'grp2'. Exiting...\n\n")
            else
                grp2[w.notValid] <- gapCode 
        }
    }


    ## Check usePairs to make sure i < j, there are no dups, etc.
    ## Then use it to create 'posTriples', the sets of positions
    ## (i,j,k) to be examined.  
    Triples <- create.posTriples(usePairs, Npos)


    ## Get consensus seq(s) for positions being used in groups
    consensus.fun <- 
        function(x) {
            tbl <- table(x)
            names(tbl)[ which.max(tbl) ]
        }

    ConsensusSeq1 <- apply(grp1, 1, consensus.fun)
    if (twoGroups)
        ConsensusSeq2 <- apply(grp2, 1, consensus.fun)



    ## Convert grp1/grp2 into integer matrix using a row-specific encoding.
    ## Each row gets encoded separately, so same codes represent different
    ## amino acids in diff. rows
    encod <- function(x) as.integer(factor(x))

    g1 <- t(apply(grp1, 1, encod))
    if (twoGroups) 
        g2 <- t(apply(grp2, 1, encod))


    ## clean up
    rm(grp1, grp2, consensus.fun, encod)
    gc()


    ## Now run the analysis on the group(s)
    if (!twoGroups) 
    {
        # - Analysis of grp1: 
        out <- singleGrpTriples(g1, Triples$posTriples, grp1.min.nc, Niter)

    } else
    {
        # - comparison of Grp1 and Grp2:
        out <- compareGrpsTriples(g1, g2, Triples$posTriples, grp1.min.nc, 
                                  grp2.min.nc, Niter)
    }



    ## ---------------------------------------- ##
    ## Make some additions to the output object ##
    ## ---------------------------------------- ##

    ## Add info from the function invocation
    out <- c(list(Call = list(call = match.call(),
                         Defaults = formals())), out)


    ## Info on how many pairs were used as the basis for the position
    ## triples, and how many duplicate triples were removed
    out$Data$Analyzed$N.usePairs <- Triples$N.usePairs
    out$Data$Analyzed$dupTriplesRemoved <- Triples$dupTriplesRemoved


    ## Add consensus seq info to output
    if (!twoGroups)
    {
        w.pos <- out$Data$Analyzed$posList
        out$Data$Analyzed$ConsensusSeq <- ConsensusSeq1[w.pos]
    } else
    {
        w.pos <- out$Data$Analyzed$posList
        out$Data$Analyzed$ConsensusSeq <- list(Grp1=ConsensusSeq1[w.pos],
                                               Grp2=ConsensusSeq2[w.pos])
    }


    ## get adjusted pvalues
    out$Results <- transform(out$Results, 
                             Mp.bon = p.adjust(Mp.raw, method="bonf"),
                             Mp.fdr = p.adjust(Mp.raw, method="fdr"),
                             Mstarp.bon = p.adjust(Mstarp.raw, method="bonf"),
                             Mstarp.fdr = p.adjust(Mstarp.raw, method="fdr"))


    ## add info on numbers of pairs significant by various criteria
    out$Signif <- list(Mp05.fdr = sum(out$Results$Mp.fdr < .05),
                       Mp10.fdr = sum(out$Results$Mp.fdr < .10),
                       Mp05.bon = sum(out$Results$Mp.bon < .05),
                       Mp10.bon = sum(out$Results$Mp.bon < .10),
                       Mstarp05.fdr = sum(out$Results$Mstarp.fdr < .05),
                       Mstarp10.fdr = sum(out$Results$Mstarp.fdr < .10),
                       Mstarp05.bon = sum(out$Results$Mstarp.bon < .05),
                       Mstarp10.bon = sum(out$Results$Mstarp.bon < .10))

    return(out)
}




singleGrpTriples <- function(g, posTriples, min.nc, Nperm)
{

    ## Number of sequences in data
    Nseq <- ncol(g)
    
    ## Calculate the marginal counts for each row.
    ## This will produce a list as pos. will have diff length marginals
    marg.cnts <- apply(g, 1, tabulate)

    ## N.c  = number of occurrances of "consensus" (most common) value
    ## N.nc = number of occurrances of non-"consensus" values
    N.c <- unlist(lapply(marg.cnts, max))
    N.nc <- Nseq - N.c

    ## Figure out which (if any) rows should be excluded from the analysis.
    ## We exclude rows that are too homogeneous (i.e. for which most seqs
    ## have the same value).
    w.excl <- which(N.nc < min.nc)
    Nexcl <- length(w.excl)


    ## Store original number of position Triples, to provide info in output obj.
    orig.Ntriples <- nrow(posTriples)


    ## if any positions don't meet criteria, they are excluded, then we
    ## do some housekeeping
    if (Nexcl > 0) 
    {
        cat("\n\nExcluded ", Nexcl, " positions from analysis due to",
            " insufficient variability.\n", nrow(g) - Nexcl,
            " positions remain.\n\n", sep="")

        ## Now that we've excluded do a last check to be sure we have
        ## enough positions remaining to run analysis.
        if (nrow(g) - Nexcl < 3)
            stop("\nMust have at least 3 positions to run analysis;\n",
                 "fewer than that remain.  Exiting...\n\n")


        ## update posTriples by removing triples involving excluded positions
        ## and then recoding positions back to 1,2,3,...
        update <- update.posTriples(w.excl, posTriples, Npos=nrow(g))

        ## updated posPairs
        posTriples <- update$posTriples

        ## retain index to original position values
        origPos <- update$origPos

        excluded.pos <- update$excluded

        g <- g[ -excluded.pos, , drop=FALSE]
        marg.cnts <- marg.cnts[ -excluded.pos ]

        rm(update) 
    } 

    ## Number of triples of positions (i,j,k) to be examined
    Ntriples <- nrow(posTriples)

    ## Number of positions in data set that will be used
    Npos <- nrow(g)

    ## Marginal probability for the levels present at each position.
    ## calculated from marginal counts 
    marg.p <- lapply(marg.cnts, function(cnts, N) cnts/N , N=Nseq)

    ## Nlev gives the number of "levels" (unique values) for each pos.
    Nlev <- unlist(lapply(marg.p, length))


    rm(marg.cnts, N.c, N.nc, w.excl) 
    gc()
    
    ## ********************************************************************** ##  
    ## Note regarding matrices being passed to C code:
    ##   - Since all objects must be passed as vectors to C, any objects we
    ##     wish to reconstruct into matrices inside the C code must be
    ##     transposed before vectorization via as.integer, as.double, etc.
    ##     This is because the vectorization is done "by column" and the 
    ##     resulting vector can't be reproduced as a matrix of the same dim. 
    ##     of the original R object.
    ## ********************************************************************** ##  

    t0 <- proc.time()[3]
    out <- .C("singleGroupTriples", 
              g = as.integer(t(g)), 
              Nseq = as.integer(Nseq),
              Npos = as.integer(Npos),
              Nlev = as.integer(Nlev),
              posI = as.integer(posTriples[,"i"]),
              posJ = as.integer(posTriples[,"j"]),
              posK = as.integer(posTriples[,"k"]),
              Ntriples = as.integer(Ntriples),

              p = as.double(unlist(marg.p)), 
              Nperm = as.integer(Nperm), 

              ## values below are calculated in program
              M = as.double(numeric(Ntriples)),
              Mp.raw = as.double(numeric(Ntriples)),
              Mstar = as.double(numeric(Ntriples)),
              Mstarp.raw = as.double(numeric(Ntriples))
             ) 

    t1 <- proc.time()[3]
    cat("Time in minutes for C code =", round((t1-t0)/60,1), "\n")


    ## construct output object to return to function 'covDoubles'
    out <- list(Analysis = list(
                    Type = "Single group: Covariation of triples of positions",
                    Min.NonCons.needed = min.nc,
                    Niter = out$Nperm
                    ),
                Data = list(
                    Analyzed = list(
                        Nseq = out$Nseq,
                        Npos = out$Npos,
                        Ntriples = out$Ntriples,
                        posList = origPos
                        ),
                    Excluded = list(
                        Npos = Nexcl,
                        Ntriples = orig.Ntriples - out$Ntriples,
                        posList = excluded.pos
                        )
                    ),
                Results = as.data.frame(
                    out[c("posI", "posJ", "posK", "M", "Mp.raw",
                          "Mstar", "Mstarp.raw")]
                    )
               )


    ## Translate positions back to original coordinates (if necessary)
    if (Nexcl > 0)
    {
        out$Results$posI <- origPos[ out$Results$posI ] 
        out$Results$posJ <- origPos[ out$Results$posJ ] 
        out$Results$posK <- origPos[ out$Results$posK ] 
    }
   
    return(out)
}



compareGrpsTriples <- function(g1, g2, posTriples, g1.min.nc, g2.min.nc, Nboot)
{
    ## Number of sequences in data
    N1seq <- ncol(g1)
    N2seq <- ncol(g2)

    ## Calculate the marginal counts for each row (Note: output is a list)
    g1.marg.cnts <- apply(g1, 1, tabulate)
    g2.marg.cnts <- apply(g2, 1, tabulate)

    ## N.c  = number of occurrances of "consensus" (most common) value
    ## N.nc = number of occurrances of non-"consensus" values
    N1.c <- unlist(lapply(g1.marg.cnts, max))
    N1.nc <- N1seq - N1.c
    N2.c <- unlist(lapply(g2.marg.cnts, max))
    N2.nc <- N2seq - N2.c

    ## Figure out which (if any) rows should be excluded from the analysis.
    ## We exclude rows that are too homogeneous (i.e. for which most seqs
    ## have the same value).
    w.excl.g1 <- which(N1.nc < g1.min.nc)
    w.excl.g2 <- which(N2.nc < g2.min.nc)

    ## Exclude positions found on *either* of the groups' exclude lists
    w.excl <- unique( c(w.excl.g1, w.excl.g2) )
    Nexcl <- length(w.excl)

    ## Store original number of position Triples, to provide info in output obj.
    orig.Ntriples <- nrow(posTriples)


    ## If there are positions to be excluded we do so, then clean up the mess
    if (Nexcl > 0)
    {
        cat("\n\nExcluded ", Nexcl, " positions from analysis due to",
            " insufficient variability.\n", nrow(g1) - Nexcl,
            " positions remain.\n\n", sep="")

        ## Now that we've excluded do a last check to be sure we have
        ## enough positions remaining to run analysis.
        if (nrow(g1) - Nexcl < 3)
            stop("\nMust have at least 3 positions to run analysis;\n",
                 "fewer than that remain.  Exiting...\n\n")


        ## update posPairs by removing pairs involving excluded positions
        ## and then recoding positions back to 1,2,3,...
        update <- update.posTriples(w.excl, posTriples, Npos = nrow(g1))

        ## updated posTriples
        posTriples <- update$posTriples

        ## retain index to original position values
        origPos <- update$origPos

        excluded.pos <- update$excluded

        ## remove excluded positions from data 
        g1 <- g1[ -excluded.pos, ]
        g2 <- g2[ -excluded.pos, ]
        g1.marg.cnts <- g1.marg.cnts[ -excluded.pos ]
        g2.marg.cnts <- g2.marg.cnts[ -excluded.pos ]

        rm(update)
    }

    ## Number of triples of positions (i,j,k) to be examined
    Ntriples <- nrow(posTriples)

    ## Number of positions in data set that will be used
    Npos <- nrow(g1)

    ## Compute marginal probability from marginal counts
    g1.marg.p <- lapply(g1.marg.cnts, function(cnts, N) cnts/N , N=N1seq)
    g2.marg.p <- lapply(g2.marg.cnts, function(cnts, N) cnts/N , N=N2seq)

    ## Nlev gives the number of "levels" (unique values) for each pos.
    N1lev <- unlist(lapply(g1.marg.p, length))
    N2lev <- unlist(lapply(g2.marg.p, length))

    rm(g1.marg.cnts, g2.marg.cnts, N1.c, N1.nc, N2.c, N2.nc,
       w.excl.g1, w.excl.g2, w.excl)
    gc()

    t0 <- proc.time()[3]
    out <- .C("compareGroupsTriples", 
               g1 = as.integer(t(g1)), 
               g2 = as.integer(t(g2)), 
               N1seq = as.integer(N1seq),
               N2seq = as.integer(N2seq),
               Npos = as.integer(Npos),
               N1lev = as.integer(N1lev),
               N2lev = as.integer(N2lev),
               posI = as.integer(posTriples[,"i"]),
               posJ = as.integer(posTriples[,"j"]),
               posK = as.integer(posTriples[,"k"]),
               Ntriples = as.integer(Ntriples),

               p1 = as.double(unlist(g1.marg.p)), 
               p2 = as.double(unlist(g2.marg.p)), 
               Nboot = as.integer(Nboot), 

               ## values below are calculated in program
               N.samps = integer(Ntriples),
               M = as.double(numeric(Ntriples)),
               Mp.raw = as.double(numeric(Ntriples)),
               Mstar = as.double(numeric(Ntriples)),
               Mstarp.raw = as.double(numeric(Ntriples))
              ) 

    t1 <- proc.time()[3]
    cat("Time in minutes for C code =", round( (t1-t0)/60, 1), "\n")


    ## construct output object to return to function 'covDoubles'
    out <- list(Analysis = list(
                    Type = "Two group comparison: Diff. in Covar. of triples of positions",
                    Min.NonCons.needed = c(Grp1 = g1.min.nc, Grp2=g2.min.nc),
                    Niter = out$Nboot
                    ),
                Data = list(
                    Analyzed = list(
                        Nseq = c(Grp1=out$N1seq, Grp2=out$N2seq),
                        Npos = out$Npos,
                        Ntriples = out$Ntriples,
                        posList = origPos
                        ),
                    Excluded = list(
                        Npos = Nexcl,
                        Ntriples = orig.Ntriples - out$Ntriples,
                        posList = excluded.pos
                        )
                    ),
                Results = as.data.frame(
                    out[c("posI", "posJ", "posK", "N.samps", "M", "Mp.raw",
                          "Mstar", "Mstarp.raw")]
                    )
               )


    ## Translate positions back to original coordinates (if necessary)
    if (Nexcl > 0)
    {
        out$Results$posI <- origPos[ out$Results$posI ]
        out$Results$posJ <- origPos[ out$Results$posJ ]
        out$Results$posK <- origPos[ out$Results$posK ]
    }

    return(out)
}

# q(save="no")
