prepare.posPairs <- function(what, posPairs, Npos)
{
    ## If the user has not provided a value for 'posPairs', a matrix of
    ## pairs of positions to be tested for covariation, then we define
    ## such an object here to be set of all unique pairs (i,j) such 
    ## that i<j , with i ranging from 1 to Npos-1 and j from 2 to Npos

    if (what == "create" && is.null(posPairs))
    {
        ## If the user has not provided a value for 'posPairs', a matrix 
        ## matrix of pairs of positions to be tested for covariation, 
        ## then we define such an object here to be set of all unique
        ## pairs (i,j) such that i<j , with 1 <= i < j <= Npos

        posPairs <- cbind(i = rep(1:Npos, each=Npos),
                          j = rep(1:Npos, times=Npos))

        i.less.than.j <- which( posPairs[,"i"] < posPairs[,"j"] )
        posPairs <- posPairs[ i.less.than.j, ]

        return(list(posPairs=posPairs, anyToDrop=FALSE, dropPos=NULL))

    } else 

        if (what == "process" && !is.null(posPairs))
    {

        ## define the vector of valid position values
        valid.positions <- 1:Npos

        ## get all unique positions involved in the pairs
        uniq.pos <- sort(unique(as.vector(posPairs)))

        ## make sure all positions in posPairs are valid (i.e. have
        ## values in the set { 1, ..., Npos}
        if ( !all(uniq.pos %in% valid.positions) )
            stop("matrix 'posPairs' contains values < 1 or > nrow(grp1)")


        ## order so col 1 < col 2 (ie. "i < j")
        posPairs <- t( apply(posPairs, 1, sort) )

        ## remove any duplicate pairs
        posPairs <- unique(posPairs)

        ## apply names to columns
        colnames(posPairs) <- c("i", "j")

        ## order rows (for aethetics of output only)
        ord <-  order(posPairs[,"i"], posPairs[,"j"])
        posPairs <- posPairs[ ord, , drop=FALSE]


        ## Now... we may have the case where not all of the positions present
        ## in the data are in the pairs of interest.  If this is the case,
        ## then (since there could be a *lot* not of interest), we will want
        ## to drop those position (rows).
        w.not.used <- which( !(valid.positions %in% uniq.pos) ) 

        ## If some pos. are not used, issue warning in case this was a mistake
        if ( length(w.not.used) > 0 )
            cat("\n\nWarning:\n", "The posPairs provided do not use ", 
                length(w.not.used), " of the positions in the data.\n",
                "The unspecified positions will be excluded.\n\n", sep="")

        return(list(posPairs = posPairs,  anyToDrop = (length(w.not.used)>0),
                    dropPos = w.not.used))

    } else

        stop("Something is messed up in function 'prepare.posPairs'\n")
}



update.posPairs <- function(action, dropPos, posPairs, Npos)
{
    ## Argument 'action' must be one of: "dropThenRecode","RecodeOnly"

    ## remove pairs involving positions in 'dropPos' (if any)
    if (action == "dropThenRecode")
    {
        ## find pairs involving dropped rows
        w.pairs.drop <- which(apply(posPairs, 1, function(x, D)
                                    any(x %in% D), dropPos) )

        posPairs <- posPairs[-w.pairs.drop, , drop=FALSE]

        ## More to drop?  In the process of dropping pairs in 'dropPos' we may
        ## have removed all occurances of other positions as well (if they only
        ## occurred in conjunction with 'dropPos' positions).  If so we should
        ## remove those pos. from the data as well.  To take account of this, we
        ## generate a list of all positions not used as the list of pos. to drop 
        uniq.pos <- sort(unique(as.vector(posPairs)))
        w.not.used <- (1:Npos)[-uniq.pos]

        ## Figure out if any more were found to be dropped
        Nextra <- length(w.not.used) - length(dropPos)

        if ( Nextra > 0 )
        {
            cat( Nextra, "more positions have been dropped.\n",
                "These positions were only present in 'posPairs' alongside\n",
                "positions dropped due to insufficient variability\n\n.") 

            if (length(uniq.pos) < 2)
                stop("Too few positions remain to perform analysis.  Exiting\n")
        }
        origPos <- uniq.pos

    } else
    {
        origPos <- (1:Npos)[-dropPos]
        w.not.used <- dropPos
    }

    posPairs[] <- match(posPairs, origPos)

    return(list(posPairs = posPairs, excluded = w.not.used, origPos = origPos))
}
