create.posTriples <- function(usePairs, Npos)
{

    ## Check here to ensure Npos >= 3. We need to have at least
    ## three positions in the data to look at triplets :)
    if ( !( Npos >= 3) ) 
        stop("\nInput data contains fewer than 3 positions (rows)\n",
             "Testing triplets of positions requires data on at ",
             "LEAST 3 positions\n\n") 

    ## define the vector of valid position values
    valid.positions <- 1:Npos

    ## get all unique positions involved in the pairs
    uniq.pos <- sort(unique(as.vector(usePairs)))

    ## make sure all positions in usePairs are valid (i.e. have
    ## values in the set { 1, ..., Npos}
    if ( !all(uniq.pos %in% valid.positions) )
        stop("\nmatrix 'usePairs' contains values < 1 or > nrow(grp1)\n\n")

    ## remove any positions for which i = j, since we need unique positions
    usePairs <- usePairs[ usePairs[,1] != usePairs[,2],  ]

    ## order so col 1 < col 2 (ie. "i < j")
    usePairs <- t( apply(usePairs, 1, sort) )

    ## remove any duplicate pairs
    usePairs <- unique(usePairs)

    ## add names to columns
    colnames(usePairs) <- c("i", "j")

    ## order rows by (i,j)
    ord <-  order(usePairs[,"i"], usePairs[,"j"])
    usePairs <- usePairs[ ord , , drop=FALSE ]


    ## Use 'usePairs' to construct 'posTriples', a 3 column matrix in
    ## which the rows define the set of position triples (i, j, k) to
    ## be tested in the program.
    posTriples <- lapply(1:nrow(usePairs), 
                         function(indx, uP, Np) {  
                             i <- uP[indx, 1] 
                             j <- uP[indx, 2] 
                             cbind( i = rep(i, Np-2),
                                    j = rep(j, Np-2),
                                    k = (1:Np)[-c(i,j)])
                         }, uP=usePairs, Np=Npos)

    ## rbind together into a single matrix
    posTriples <- do.call("rbind", posTriples)



    # Now remove any duplicate triples that may exist in posTriples.
    # This is done to: (a) reduce computation time, and (b) to allow
    # for correct pvalue adjustment; and is done because there doesn't
    # seem to be any good reason for keeping the duplicate triples. 
    # 
    # How duplicate triples can occur (an example): 
    # ---------------------------------------------
    # Assume 'usePairs' contains (1, 10) and (5, 10), then posTriples 
    # (as constructed) will contain ((1, 10), 5) and ((5, 10), 1),
    # which are (of course) the same, as ordering is not important. 
     
    # (1) create version of 'posTriples' with columns ordered i < j < k
    posTrip.ordCol <- t( apply(posTriples, 1, sort) )

    # (2) find dup. rows and remove them (first occurrance is kept)
    isDuplicated <- duplicated(posTrip.ordCol)
    posTriples <- posTriples[ !isDuplicated, , drop=FALSE ]

    list(posTriples = posTriples,
         dupTriplesRemoved = sum(isDuplicated),
         N.usePairs = nrow(usePairs))
}



update.posTriples <- function(dropPos, posTriples, Npos)
{
    ## remove triples involving positions in 'dropPos' (if any)

    ## find triples involving positions we wish to drop
    w.triples.drop <- which(apply(posTriples, 1, function(x, D)
                                    any(x %in% D), dropPos) )

    ## drop those triples
    posTriples <- posTriples[ -w.triples.drop , , drop=FALSE]


    ## Check to make sure we still have any triples left - it's 
    ## entirely possible that we don't!
    if ( nrow(posTriples) == 0 )
        stop ("\nAll pairs specified in 'usePairs' contain positions\n",
              "dropped due to insufficient variation, hence no\n",
              "triples remain for testing.  Exiting...\n\n")


    ## *** NOTE:  ********************************************************
    ## ***   next few lines aren't needed unless the program is extended
    ## ***   to allow direct specification of triples to test.  However,
    ## ***   it does no harm to leave them here...
    ## *******************************************************************
    ## More to drop?  In the process of dropping pairs in 'dropPos' we may
    ## have removed all occurances of other positions as well (if they only
    ## occurred in conjunction with 'dropPos' positions).  If so we should
    ## remove those pos. from the data as well.  To take account of this, we
    ## generate a list of all positions not used as the list of pos. to drop 
    uniq.pos <- sort(unique(as.vector(posTriples)))
    w.not.used <- (1:Npos)[-uniq.pos]

    origPos <- uniq.pos
    posTriples[] <- match(posTriples, origPos)

    return(list(posTriples = posTriples, excluded = w.not.used,
                origPos = origPos))
}
